SIMPLE_MAP = "simple_map"
LARGE_MAP = "large_map"
RESOLUTION = 0.01

ROBOT_RADIUS = 0.3
MAX_LINEAR_VELOCITY = 1.0
MIN_LINEAR_VELOCITY = 0.0
MAX_ANGULAR_VELOCITY = 1.0
MIN_ANGULAR_VELOCITY = -MAX_ANGULAR_VELOCITY

DT = 0.1
MAX_STEPS = 500

MIN_RANGE = 0.0
MAX_RANGE = 10.0
NUM_LIDAR = 720
MAX_ANGLE = 1.5707963267948966
ANGLE_INCREMENT = MAX_ANGLE*2.0/NUM_LIDAR

MIN_TARGET_DISTANCE = 0.0
MAX_TARGET_DISTANCE = 20.0
